# -*-coding: utf-8 -*-
"""
    @Author : PKing
    @E-mail : 390737991@qq.com
    @Date   : 2023-03-28 09:44:09
    @Brief  : Pytorch可视化神经网络热力图(CAM):
              https://blog.csdn.net/sinat_37532065/article/details/103362517
              https://github.com/frgfm/torch-cam
              https://zhuanlan.zhihu.com/p/479485138
"""
import os
import torch
import numpy as np
import cv2
import matplotlib.pyplot as plt
from pybaseutils import image_utils, file_utils


class HeatMap(object):
    """绘制类激活映射"""

    def __init__(self, flag="hm", use_grad=True):
        """
        Grad-CAM类激活映射可视化:
        :param flag:
        :param use_grad:
        """
        self.flag = flag
        self.use_grad = use_grad
        self.images = []
        self.names = []
        self.outputs = None
        self.features = None

    def set_image(self, image, name=None):
        if not name: name = f"{file_utils.get_time()}.jpg"
        if not isinstance(image, list): image = [image]
        if not isinstance(name, list): name = [name]
        self.images = image
        self.names = name

    def set_outputs(self, outputs: torch.Tensor):
        self.outputs = outputs

    def set_features(self, features: torch.Tensor):
        self.features = features

    def get_heatmap(self, vis=False, out=""):
        """
        获得热力图可视化结果
        :param vis: 是否显示
        :param out: 是否保存热力图可视化图
        :return: featuremaps： 返回featuremap特征图
                 heatmaps   ： 返回heatmaps热力图
        """
        print(f"features:{self.features.shape}, outputs:{self.outputs.shape}")
        featuremaps, heatmaps = draw_heatmap_from_output(self.outputs, self.features, images=self.images,
                                                         use_grad=self.use_grad, vis=vis)
        if out:
            self.save_heatmap(featuremaps, heatmaps, out)
        return featuremaps, heatmaps

    def save_heatmap(self, featuremaps, heatmaps, out: str, combine=True):
        if not out: return
        for i in range(len(featuremaps)):
            name = os.path.basename(self.names[i]).split(".")[0]
            if combine:
                file = file_utils.create_dir(out, None, f"{name}-{self.flag}.jpg")
                image = image_utils.image_hstack([featuremaps[i], heatmaps[i]])
                cv2.imwrite(file, image)
            else:
                hfile = file_utils.create_dir(out, None, f"{name}-{self.flag}-featuremap.jpg")
                vfile = file_utils.create_dir(out, None, f"{name}-{self.flag}-heatmap-image.jpg")
                cv2.imwrite(hfile, featuremaps[i])
                cv2.imwrite(vfile, heatmaps[i])


def draw_heatmap_from_output(output: torch.Tensor, features: torch.Tensor, images, use_grad=True, alpha=0.3, vis=False):
    """
    绘制 Class Activation Map
    :param output: 模型输出结果
    :param features: 模型输出特征图
    :param images: 原始图片
    :param use_grad: 是否考虑梯度
    :param vis: 是否可视化原始heatmap（调用matplotlib）
    :return:
    """
    if not isinstance(images, list): images = [images]

    # 为了能读取到中间梯度定义的辅助函数
    def extract(g):
        global features_grad
        features_grad = g

    # 预测得分最高的那一类对应的输出score
    index = torch.argmax(output).item()
    pred_class = output[:, index].mean()
    if use_grad:
        features.register_hook(extract)
        pred_class.backward()  # 计算梯度
        grads = features_grad  # 获取梯度
        avg_grads = torch.nn.functional.adaptive_avg_pool2d(grads, (1, 1))

    featuremaps, heatmaps = [], []
    for b in range(features.size(0)):
        feature = features[b]
        if use_grad:
            avg_grad = avg_grads[b]
            for i in range(feature.size(0)):  feature[i, ...] *= avg_grad[i, ...]
        f, h = plot_heatmap(feature.cpu().detach().numpy(), images[b], alpha=alpha, vis=vis)
        featuremaps.append(f)
        heatmaps.append(h)
    return featuremaps, heatmaps


def draw_heatmap_from_model(model, inputs_tensor, image, use_grad=True, vis=False):
    """
    :param model:
    :param inputs_tensor:
    :param image:
    :param use_grad:
    :param vis:
    :return:
    """
    # 获取模型输出的feature/score
    model.eval()
    scores = model(inputs_tensor)
    features = model.features
    featuremaps, heatmaps = draw_heatmap_from_output(features, scores, images=image, use_grad=use_grad, vis=vis)
    return featuremaps, heatmaps


def plot_heatmap(feature, image, alpha=0.3, vis=False):
    feature = np.mean(feature, axis=0)

    feature = np.maximum(feature, 0)
    feature /= np.max(feature)
    # feature=image_utils.cv_image_normalization(feature,0,1.0)

    # CAM结果保存路径
    H, W = image.shape[:2]
    f = cv2.resize(feature, dsize=(W, H), interpolation=cv2.INTER_LINEAR)  # 将热力图的大小调整为与原始图像相同
    f = np.uint8(np.clip(255 * f, 0, 255))  # 将热力图转换为RGB格式
    f = cv2.applyColorMap(f, cv2.COLORMAP_JET)  # 将热力图应用于原始图像
    heatmap = np.float32(f) * (1 - alpha) + np.float32(image) * alpha  # 这里的0.4是热力图强度因子
    heatmap = np.uint8(np.clip(heatmap, 0, 255))  # 将热力图转换为RGB格式

    fig = plt.matshow(feature)
    featuremap = image_utils.fig2data(fig.figure)
    featuremap = cv2.cvtColor(featuremap, cv2.COLOR_RGB2BGR)
    if vis:
        # plt.show()
        image_utils.cv_show_image("featuremap", featuremap, delay=1)
        image_utils.cv_show_image("heatmap", heatmap, delay=10)
    return featuremap, heatmap



# -*- coding: utf-8 -*-
"""
    @Author : Pan
    @Brief  : NCNN inference engine
"""
import os
import sys
import ncnn
import numpy as np


class NCNNEngine(object):
    def __init__(self, par_file, bin_file="", num_thread=8, use_gpu=False, use_fp16=False):
        """
        NCNN推理引擎，需要注意：
        (1)将输入数组转换为C/C++数组时，需要保证输入数据是顺序（行优先）的连续数组，否则会出现推理异常等问题
           为避免内存不连续，建议使用np.ascontiguousarray(data)或者使用copy()
        (2)NCNN仅仅支持(c,h,w),不能输入(1,c,h,w),否则推理结果异常，建议使用np.squeeze(data)简化数组维度
        :param par_file: NCNN param file path (.param)
        :param bin_file: NCNN model weights file path (.bin)
        :param num_thread: CPU线程个数，-1表示等于系统cpu个数
        :param use_gpu: Whether to use GPU (Vulkan) for inference
        :param use_fp16: Whether to enable FP16 inference
        pip install ncnn
        """
        if not bin_file: bin_file = par_file.replace(".param", ".bin")
        assert os.path.exists(par_file), f"*.param file not exists:{par_file}"
        assert os.path.exists(bin_file), f"*.bin   file not exists:{bin_file}"
        self.cpu_info = ncnn.get_cpu_count()
        self.gpu_info = ncnn.get_gpu_device()
        self.num_thread = num_thread if num_thread > 0 else self.cpu_info
        # TODO 创建 ncnn 的 Net 对象
        self.net = ncnn.Net()
        self.net.opt.num_threads = self.num_thread  # must set net.opt.num_threads=N before net.load_param()
        if use_gpu:  # Enable vulkan compute if GPU is requested
            self.net.opt.use_vulkan_compute = True
            if use_fp16:
                self.net.opt.use_fp16_packed = True
                self.net.opt.use_fp16_storage = True
                self.net.opt.use_fp16_arithmetic = True
        # Load model
        self.net.load_param(par_file)
        self.net.load_model(bin_file)
        # Get input and output names from net
        self.inp_names = self.net.input_names()
        self.out_names = self.net.output_names()
        print("cpu_info          :{},use threads:{}".format(self.cpu_info, self.num_thread))
        print("gpu_info          :{}".format(self.gpu_info))
        print("use_gpu           :{}".format(use_gpu))
        print("use_fp16          :{}".format(use_fp16))
        print("inp_names         :{}".format(self.inp_names))
        print("out_names         :{}".format(self.out_names))
        print("param_file        :{}".format(par_file))
        print("bin_file          :{}".format(bin_file))
        print('-----------' * 5, flush=True)

    def __call__(self, image_tensor):
        """
        :param image_tensor: numpy array with shape (b, c, h, w)
        :return: model outputs
        """
        outputs = self.forward(image_tensor)
        return outputs

    def forward(self, image_tensor):
        """
        :param image_tensor: numpy array with shape (b, c, h, w)
        :return: list of outputs
        """
        with self.net.create_extractor() as ex:
            # ex.set_num_threads(self.num_thread);
            # BUG：为避免内存不连续，需要copy或者np.ascontiguousarray()
            data = np.ascontiguousarray(np.squeeze(image_tensor))  # 使用np.squeeze(data)简化数组维度，避免结果异常
            # data = np.ascontiguousarray(image_tensor)    # BUG: 这样写，推理结果有异常，没有squeeze
            # input = ncnn.Mat(np.ascontiguousarray(data)) # BUG: 这样写，推理结果有异常
            input = ncnn.Mat(data)  # 推理结果正常  
            ex.input(self.inp_names[0], input)
            # Extract outputs
            outputs = []
            for name in self.out_names:
                _, out = ex.extract(name)
                outputs.append(np.array(out).reshape(1, out.c, out.h, out.w))
        return outputs

    def forward_image(self, image, mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5]):
        """
        :param image: numpy array with shape (h, w,3)
        :return: list of outputs
        """
        mean = np.array(mean, dtype=np.float32) * 255
        std = 1 / np.array(std, dtype=np.float32) / 255
        with self.net.create_extractor() as ex:
            # Convert numpy array to ncnn.Mat
            data = np.ascontiguousarray(np.squeeze(image))  # BUG：内存不连续，需要copy或者np.ascontiguousarray()
            input = ncnn.Mat.from_pixels(data, ncnn.Mat.PixelType.PIXEL_RGB, data.shape[1], data.shape[0])  # (w,h)
            input.substract_mean_normalize(mean, std);
            ex.input(self.inp_names[0], input.clone())
            # Extract outputs
            outputs = []
            for name in self.out_names:
                _, out = ex.extract(name)
                outputs.append(np.array(out).reshape(1, out.c, out.h, out.w))
        return outputs

    def performance(self, inputs, iterate=20):
        """
        :param inputs: inputs (1,c,h,w)
        :param iterate: number of iterations for performance testing
        :return: outputs
        """
        from pybaseutils import time_utils
        outputs = self.forward(inputs)
        for i in range(iterate):
            with time_utils.Performance() as p:
                outputs = self.forward(inputs)
        print("inp=\n", inputs[0, 0, 0, 0:10])
        print("out=")
        [print(out[0, 0, 0, 0:10]) for out in outputs]
        return outputs


if __name__ == "__main__":
    np.random.seed(100)
    # Example usage
    param_file = "../../data/model/ncnn-fp16/model.ncnn.param"
    # param_file = "../../data/model/ncnn-fp32/model.ncnn.param"
    # Create random input
    input_size = [168, 168]
    inputs = np.random.random(size=(1, 3, input_size[1], input_size[0]))
    inputs = np.asarray(inputs, dtype=np.float32)
    model = NCNNEngine(param_file, use_gpu=True, use_fp16=True)
    model.performance(inputs)
    print("----")



# -*-coding: utf-8 -*-
"""
    @Author : PKing
    @E-mail : 390737991@qq.com
    @Date   : 2023-03-29 20:09:27
    @Brief  : Pytorch可视化神经网络热力图(CAM):
              https://blog.csdn.net/sinat_37532065/article/details/103362517
              https://github.com/frgfm/torch-cam
              https://zhuanlan.zhihu.com/p/479485138
"""
import os
import sys

sys.path.insert(0, os.getcwd())
import cv2
import torch
import numpy as np
import matplotlib.pyplot as plt
import PIL.Image as Image
import torchcam
from pybaseutils import file_utils, image_utils
from torchcam import methods

CAM_TYPE = ["CAM", "GradCAM", "GradCAMpp", "SmoothGradCAMpp", "XGradCAM", "LayerCAM"]


class TorchCAM(object):
    """绘制类激活映射"""

    def __init__(self, model: torch.nn.Module, target_layer=None, input_shape=(3, 224, 224),
                 cam_type="GradCAM", flag="hm"):
        """
        类激活映射: https://github.com/frgfm/torch-cam
        :param model: 模型
        :param target_layer: 可视化指定层的名称
        :param input_shape: 模型输入维度(C,H,W)
        :param cam_type: 可视化方法： ["CAM", "GradCAM", "GradCAMpp", "SmoothGradCAMpp", "XGradCAM", "LayerCAM"]
        :param flag:
        """
        self.flag = flag
        self.images = []
        self.names = []
        self.outputs = None
        self.features = None
        self.model = model
        if len(input_shape) == 2:
            input_shape = (3, input_shape[1], input_shape[0])
        elif len(input_shape) == 4:
            b, c, h, w = input_shape
            input_shape = (c, h, w)
        self.cam = self.build_cam(cam_type=cam_type, target_layer=target_layer, input_shape=input_shape)
        self.target_layer = self.cam.target_names
        if target_layer is None:
            print(f"target_layer is not specified, automatic resolution target_layer is {self.target_layer}")
        else:
            print(f"target_layer is {self.target_layer}")

    def build_cam(self, cam_type="GradCAM", target_layer=None, input_shape=(3, 224, 224)):
        """
        :param cam_type: ["GradCAM", "GradCAMpp", "SmoothGradCAMpp", "XGradCAM", "LayerCAM"]
        :param target_layer:
        :param input_shape:
        :return:
        """
        if cam_type.lower() == "CAM".lower():
            cam = methods.CAM(self.model, target_layer=target_layer, input_shape=input_shape)
        elif cam_type.lower() == "GradCAM".lower():
            cam = methods.GradCAM(self.model, target_layer=target_layer, input_shape=input_shape)
        elif cam_type.lower() == "GradCAMpp".lower():
            cam = methods.GradCAMpp(self.model, target_layer=target_layer, input_shape=input_shape)
        elif cam_type.lower() == "SmoothGradCAMpp".lower():
            cam = methods.SmoothGradCAMpp(self.model, target_layer=target_layer, input_shape=input_shape)
        elif cam_type.lower() == "XGradCAM".lower():
            cam = methods.XGradCAM(self.model, target_layer=target_layer, input_shape=input_shape)
        elif cam_type.lower() == "LayerCAM".lower():
            cam = methods.LayerCAM(self.model, target_layer=target_layer, input_shape=input_shape)
        else:
            cam = None
        return cam

    def set_image(self, image, name=None):
        if not name: name = f"{file_utils.get_time()}.jpg"
        if not isinstance(image, list): image = [image]
        if not isinstance(name, list): name = [name]
        self.images = image
        self.names = name

    def set_outputs(self, outputs: torch.Tensor):
        self.outputs = outputs

    def set_features(self, features: torch.Tensor):
        # self.features = features
        pass

    def get_heatmap(self, vis=False, out=""):
        """
        获得热力图可视化结果
        :param vis: 是否显示
        :param out: 是否保存热力图可视化图
        :return: featuremaps： 返回featuremap特征图
                 heatmaps   ： 返回heatmaps热力图
        """
        pred_index = self.outputs.argmax().item()
        outputs = self.cam(class_idx=pred_index, scores=self.outputs)
        features, = outputs
        featuremaps, heatmaps = self.draw_heatmap_from_output(features, vis=vis)
        if out:
            self.save_heatmap(featuremaps, heatmaps, out)
        return featuremaps, heatmaps

    def draw_heatmap_from_output(self, features: torch.Tensor, vis=False):
        heatmaps = []
        featuremaps = []
        features = features.cpu().detach().numpy()
        for b in range(len(features)):
            feature = features[b]
            mask = np.array(feature, dtype=np.float32)
            image = Image.fromarray(self.images[b])
            mask = Image.fromarray(mask)
            heatmap = torchcam.utils.overlay_mask(image, mask, alpha=0.3)
            heatmap = np.array(heatmap)
            heatmap = cv2.cvtColor(heatmap, cv2.COLOR_RGB2BGR)
            #
            feature = np.mean(feature, axis=0)
            feature = np.maximum(feature, 0)
            feature /= np.max(feature)
            fig = plt.matshow(features[b])
            featuremap = image_utils.fig2data(fig.figure)
            featuremap = cv2.cvtColor(featuremap, cv2.COLOR_RGB2BGR)

            featuremaps.append(featuremap)
            heatmaps.append(heatmap)
            if vis:
                image_utils.cv_show_image("featuremap", featuremap, delay=1)
                image_utils.cv_show_image("heatmap", heatmap, delay=10)
        return featuremaps, heatmaps

    def save_heatmap(self, featuremaps, heatmaps, out: str, combine=True):
        if not out: return
        for i in range(len(featuremaps)):
            name = os.path.basename(self.names[i]).split(".")[0]
            if combine:
                file = file_utils.create_dir(out, None, f"{name}-{self.flag}.jpg")
                image = image_utils.image_hstack([featuremaps[i], heatmaps[i]])
                cv2.imwrite(file, image)
            else:
                hfile = file_utils.create_dir(out, None, f"{name}-{self.flag}-featuremap.jpg")
                vfile = file_utils.create_dir(out, None, f"{name}-{self.flag}-heatmap-image.jpg")
                cv2.imwrite(hfile, featuremaps[i])
                cv2.imwrite(vfile, heatmaps[i])



# -*- coding: utf-8 -*-
"""
    @Author : PKing
    @E-mail : 
    @Date   : 2024-05-22 14:11:49
    @Brief  : https://blog.csdn.net/weixin_44613415/article/details/131850160
"""
import os
import cv2
import numpy as np
import os
import tensorrt as trt
import pycuda.driver as cuda
import ctypes
import logging
from pybaseutils import file_utils, image_utils

logger = logging.getLogger(__name__)
ctypes.pythonapi.PyCapsule_GetPointer.restype = ctypes.c_char_p
ctypes.pythonapi.PyCapsule_GetPointer.argtypes = [ctypes.py_object, ctypes.c_char_p]


def build_transform(image, input_size):
    """
    数据预处理操作
    :param image:
    :param input_size:
    :return:
    """
    image = cv2.resize(image, (input_size[1], input_size[0]))
    image = image.transpose((2, 0, 1)).astype(np.float32)
    image /= 255.0
    return image


class ImageDataset(object):
    """数据加载类用于加载校准数据"""

    def __init__(self, image_dir, transform=build_transform, input_size=(640, 640), use_rgb=True, batch_size=8,
                 epoch=10):
        """
        :param image_dir:
        :param transform:
        :param input_size:
        :param use_rgb:
        :param batch_size:
        :param epoch:
        """
        self.index = 0
        self.epoch = epoch
        self.use_rgb = use_rgb
        self.input_size = input_size
        self.batch_size = batch_size
        self.transform = transform
        self.file_list = file_utils.get_images_list(image_dir)
        self.nums = len(self.file_list)  # 图片个数
        assert self.nums > 0, Exception("图片为空：{}".format(image_dir))
        self.iteration = self.batch_size * self.epoch  # 总迭代数
        if self.nums < self.iteration:
            self.file_list = self.file_list * ((self.iteration + self.nums - 1) // self.nums)
        self.batch_data = np.zeros((self.batch_size, 3, self.input_size[1], self.input_size[0]), dtype=np.float32)
        print("images:{}/{},total iteration:{}".format(self.nums, len(self.file_list), self.iteration))
        print("epoch:{},batch_size:{},input_size :{}".format(self.epoch, self.batch_size, self.input_size))

    def reset(self):
        self.index = 0

    def next_batch(self):
        if self.index < self.epoch:
            for i in range(self.batch_size):
                image_file = self.file_list[i + self.index * self.batch_size]
                image = cv2.imread(image_file)
                if self.use_rgb: image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
                if self.transform: image = self.transform(image, input_size=self.input_size)
                self.batch_data[i] = image
            self.index += 1
            return np.ascontiguousarray(self.batch_data, dtype=np.float32)
        else:
            return np.array([])

    def __len__(self):
        return self.epoch


class Calibrator(trt.IInt8EntropyCalibrator):
    """
    校准器Calibrator,TensorRT提供的四个校准器类中的一个，需要重写父校准器的几个方法
    IInt8EntropyCalibrator2
    IInt8LegacyCalibrator
    IInt8EntropyCalibrator
    IInt8MinMaxCalibrator
    """

    def __init__(self, dataset: ImageDataset, cache_file="./calibration.cache"):
        """
        :param dataset:
        :param cache_file:
        """
        trt.IInt8EntropyCalibrator.__init__(self)
        self.dataset = dataset
        self.d_input = cuda.mem_alloc(self.dataset.batch_data.nbytes)
        self.cache_file = cache_file
        dataset.reset()

    def get_batch_size(self):
        """用于获取batch的大小"""
        return self.dataset.batch_size

    def get_batch(self, names):
        """用于获取一个batch的数据"""
        batch = self.dataset.next_batch()
        if not batch.size:
            return None
        # 把校准数据从CPU搬运到GPU中
        cuda.memcpy_htod(self.d_input, batch)
        return [int(self.d_input)]

    def read_calibration_cache(self):
        """用于从文件中读取校准表"""
        # If there is a cache, use it instead of calibrating again. Otherwise, implicitly return None.
        if os.path.exists(self.cache_file):
            with open(self.cache_file, "rb") as f:
                logger.info("Using calibration cache to save time: {:}".format(self.cache_file))
                return f.read()

    def write_calibration_cache(self, cache):
        """用于把校准表从内存中写入文件中"""
        with open(self.cache_file, "wb") as f:
            logger.info("Caching calibration data for future use: {:}".format(self.cache_file))
            f.write(cache)


if __name__ == '__main__':
    image_dir = "/home/PKing/nasdata/Project/3D/Depth-Anything/Depth-Anything/assets/test_image"
    dataset = ImageDataset(image_dir, input_size=(640, 640), use_rgb=True, batch_size=8, epoch=2)
    for i in range(dataset.epoch):
        data = dataset.next_batch()
        print(data.shape)



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2021-08-12 20:27:27
"""

import torch
import torch.optim as optim
import torch.nn as nn
import numpy as np
from ..callbacks.callbacks import Callback
from .WarmUpLR import WarmUpLR
from torch.optim import lr_scheduler


class CosineLR(Callback):
    def __init__(self,
                 optimizer,
                 epochs,
                 num_steps,
                 num_cycles=1,
                 lr_init=0.1,
                 decay=0.99,
                 num_warn_up=10,
                 ):
        """
        余弦退火学习率调整策略
        optimizer (Optimizer): Wrapped optimizer.
        t_max (int): Maximum number of iterations.
        eta_min (float): Minimum learning rate. Default: 0.
        last_epoch (int): The index of last epoch. Default: -1.
        verbose (bool): If ``True``, prints a message to stdout for each update. Default: ``False``.
        :param optimizer:
        :param epochs:
        :param num_steps: 一个epoch的迭代次数，len(self.train_dataloader)
        :param num_cycles: 周期次数
        :param lr_init: is init lr.
        :param decay: 振幅衰减率,当1.0时,每次周期循环不衰减
        :param num_warn_up:
        """
        super(CosineLR, self).__init__()
        self.num_steps = num_steps
        self.num_cycles = num_cycles
        self.num_warn_up = num_warn_up
        self.decay = decay
        self.lr_init = lr_init
        self.optimizer = optimizer
        self.epoch = 0
        self.epochs = epochs - self.num_warn_up - 1
        t_max = self.epochs * 1.0 / (2 * self.num_cycles - 1)  # 一次学习率周期的迭代次数，即 T_max 个 epoch 之后重新设置学习率。
        eta_min = 0.00001  # 最小学习率，即在一个周期中，学习率最小会下降到 eta_min，默认值为 0
        self.scheduler = lr_scheduler.CosineAnnealingLR(optimizer, t_max, eta_min=eta_min, last_epoch=-1)
        self.warm_up = WarmUpLR(optimizer,
                                num_steps=self.num_steps,
                                lr_init=lr_init,
                                num_warn_up=num_warn_up)

    def get_lr(self, epoch):
        epoch = epoch - self.num_warn_up
        self.scheduler.step(epoch)
        # self.scheduler.step()
        # self.scheduler.last_epoch =epoch
        lr = self.optimizer.param_groups[0]["lr"]
        lr = lr * self.decay ** ((self.num_cycles - 0.5) * epoch / self.epochs)
        return lr

    def set_lr(self, lr):
        for param_group in self.optimizer.param_groups:
            param_group["lr"] = lr

    def on_epoch_begin(self, epoch, logs: dict = {}):
        self.epoch = epoch
        self.set_lr(self.get_lr(epoch))

    def on_batch_end(self, batch, logs: dict = {}):
        self.step(epoch=self.epoch, step=batch)

    def step(self, epoch=0, step=0):
        # step每次迭代都会调用，比较耗时，建议与step无关的操作放在on_epoch_begin中
        # total_step = self.num_steps * epoch + step
        # self.scheduler.step(epoch)
        self.warm_up.step(epoch, step)


class CosineAnnealingLR(Callback):
    def __init__(self,
                 optimizer,
                 epochs,
                 num_steps,
                 num_cycles=1,
                 lr_init=0.1,
                 decay=0.99,
                 num_warn_up=10,
                 ):
        """
        余弦退火学习率调整策略
        optimizer (Optimizer): Wrapped optimizer.
        t_max (int): Maximum number of iterations.
        eta_min (float): Minimum learning rate. Default: 0.
        last_epoch (int): The index of last epoch. Default: -1.
        verbose (bool): If ``True``, prints a message to stdout for each update. Default: ``False``.
        :param optimizer:
        :param epochs:
        :param num_steps: 一个epoch的迭代次数，len(self.train_dataloader)
        :param num_cycles: 周期次数
        :param lr_init: is init lr.
        :param decay: 振幅衰减率,当1.0时,每次周期循环不衰减
        :param num_warn_up:
        """
        super(CosineAnnealingLR, self).__init__()
        self.num_steps = num_steps
        self.num_cycles = num_cycles
        self.num_warn_up = num_warn_up
        self.decay = decay
        self.lr_init = lr_init
        self.optimizer = optimizer
        self.epoch = 0
        self.epochs = epochs
        self.warmup = lr_scheduler.LinearLR(self.optimizer,
                                            start_factor=0.1,
                                            total_iters=self.num_warn_up  # warmup
                                            )
        self.cosine = lr_scheduler.CosineAnnealingLR(self.optimizer,
                                                     T_max=self.epochs - self.num_warn_up,
                                                     eta_min=1e-6,
                                                     last_epoch=-1
                                                     )
        self.scheduler = lr_scheduler.ChainedScheduler([self.warmup, self.cosine])

    def on_epoch_begin(self, epoch, logs: dict = {}):
        self.epoch = epoch
        # self.scheduler.step()
        if self.epoch < self.num_warn_up:
            self.scheduler._schedulers[0].step(self.epoch)
        else:
            self.scheduler._schedulers[1].step(self.epoch - self.num_warn_up)

    def on_batch_end(self, batch, logs: dict = {}):
        self.step(epoch=self.epoch, step=batch)

    def step(self, epoch=0, step=0):
        # step每次迭代都会调用，比较耗时，建议与step无关的操作放在on_epoch_begin中
        # total_step = self.num_steps * epoch + step
        # self.scheduler.step(epoch)
        pass



# -*- coding: utf-8 -*-
"""
# --------------------------------------------------------
# @Author :
# @E-mail :
# @Date   : 2020-04-03 18:38:34
# --------------------------------------------------------
"""
import sys
import os

sys.path.append(os.getcwd())
import numpy as np
from tqdm import tqdm
from basetrainer.metric.map_tools.evaluator import map_eval
from pybaseutils import file_utils, json_utils


class MapMetric(object):
    """
    使用方法见evaluate_example或者evaluate_for_json
    mAP参考资料：https://github.com/Cartucho/mAP
    """

    def __init__(self, min_iou=0.5):
        """
        :param min_iou: IOU置信度，大于该值，则认为是TP
        """
        self.min_iou = min_iou

    def evaluate(self, gt_info: dict, dt_info: dict, class_name: list, plot=False):
        """
        :param gt_info: ground-truth results,format is
               gt_info={"class_name1":
                                    {"file-name1":[[score,x1,y1,x2,y2],[score,x1,y1,x2,y2],[]...]}
                                    {"file-name2":[[score,x1,y1,x2,y2],[score,x1,y1,x2,y2],[]...]}，
                        "class_name2":
                                    {"file-name1":[[score,x1,y1,x2,y2],[score,x1,y1,x2,y2],[]...]}
                                    {"file-name2":[[score,x1,y1,x2,y2],[score,x1,y1,x2,y2],[]...]}
                         }
        :param dt_info: detection-results，format is the same with gt_info
        :param class_name: 需要测评的类别
        :param plot: 是否显示绘图结果
        :return:
        """
        ap_logs = "Average Precision Per-class:\n"
        results = {}
        for name in class_name:
            gt = gt_info[name] if name in gt_info else {}
            dt = dt_info[name] if name in dt_info else {}
            if not gt: continue
            if not dt: continue
            r = map_eval.compute_average_precision_per_class(gt, dt, self.min_iou, use_2007_metric=False, name=name)
            results[name] = r
        aps = {k: v["ap"] for k, v in results.items()}
        map = sum(aps.values()) / len(aps)
        ap_logs += "{}\n".format(aps)
        ap_logs += "Average Precision Across All Classes:\nmAP:{:3.5f}".format(map)
        print(ap_logs)
        print("n_classes={}".format(len(class_name)))
        print("==" * 20)
        if plot and map > 0:
            map_eval.plot_map_recall_precision(results)
        return aps, map

    def build_data_info(self, data_info: dict, file_name, boxes: list, label: list, score=[], class_name=[]):
        """
        构建gt_info和dt_info,使用方法见evaluate_example
        :param data_info: 输入/输出结果，用于保存所有目标数据信息
        :param file_name: 用于文件标识的唯一ID，一般用文件名称
        :param boxes: shape is (n,4),is (xmin,ymin,xmax,ymax)
        :param label: shape is (n,),type is np.int32
        :param score: shape is (n,),type is np.float32
        :param class_name:类别列表,用于转换label，class_name[int(label[i])]
        :return:
        """
        if isinstance(label, np.ndarray): label = label.tolist()
        if isinstance(boxes, np.ndarray): boxes = boxes.tolist()
        if isinstance(score, np.ndarray): score = score.tolist()
        for i in range(len(label)):
            n = class_name[int(label[i])] if class_name else label[i]
            b = boxes[i]
            if len(score) > 0:  b = [score[i]] + b
            if n not in data_info: data_info[n] = {}
            if file_name not in data_info[n]: data_info[n][file_name] = []
            data_info[n][file_name].append(b)
        return data_info

    def evaluate_example(self, plot=True):
        """
        使用例子
        :param plot:
        :return:
        """
        class_name = ['BACKGROUND', 'hand']
        dataset = ...
        gt_info = {}
        dt_info = {}
        for i in tqdm(range(len(dataset))):
            file = ...
            # ground-truth results
            gt_boxes = ...  # shape is (n,4),is (xmin,ymin,xmax,ymax)
            gt_label = ...  # shape is (n,),type is np.float32
            self.build_data_info(gt_info, file, gt_boxes, gt_label, class_name=class_name)
            # detection results
            dt_boxes = ...  # shape is (n,4),is (xmin,ymin,xmax,ymax)
            dt_score = ...  # shape is (n,),type is np.float32
            dt_label = ...  # shape is (n,),type is np.int32
            self.build_data_info(dt_info, file, dt_boxes, dt_label, dt_score, class_name=class_name)
        self.evaluate(gt_info, dt_info, class_name, plot=plot)

    def evaluate_for_json(self, dt_file="dt_info.json", gt_file="gt_info.json"):
        """
       dt_file or gt_file format is
        {"class_name1":
                    {"file-name1":[[score,x1,y1,x2,y2],[score,x1,y1,x2,y2],[]...]}
                    {"file-name2":[[score,x1,y1,x2,y2],[score,x1,y1,x2,y2],[]...]}，
        "class_name2":
                    {"file-name1":[[score,x1,y1,x2,y2],[score,x1,y1,x2,y2],[]...]}
                    {"file-name2":[[score,x1,y1,x2,y2],[score,x1,y1,x2,y2],[]...]}
         }
        :param dt_file:
        :param gt_file:
        :return:
        """
        class_name = ['BACKGROUND', 'hand']
        dt_info = json_utils.read_json_data(dt_file)
        gt_info = json_utils.read_json_data(gt_file)
        self.evaluate(gt_info, dt_info, class_name, plot=True)

    def evaluate_for_txt(self, dt_file="", gt_file=""):
        """
        mAP参考资料：https://github.com/Cartucho/mAP
        In these *.txt files, each line should be in the following format:
        class_name1 xmin ymin xmax ymax
        class_name2 xmin ymin xmax ymax
        ...
        :param dt_file:
        :param gt_file:
        :return:
        """
        dt_file = "/home/PKing/nasdata/Project/mAP/input/detection-results"
        gt_file = "/home/PKing/nasdata/Project/mAP/input/ground-truth"
        dt_files = file_utils.get_files_list(dt_file, postfix=["*.txt"])
        gt_files = file_utils.get_files_list(gt_file, postfix=["*.txt"])
        dt_info = {}
        for file in dt_files:
            data_info = file_utils.read_data(file, split=" ")
            name = [info[0] for info in data_info]
            score = [info[1] for info in data_info]
            boxes = [info[2:] for info in data_info]
            self.build_data_info(dt_info, os.path.basename(file), boxes, name, score=score)
        class_name = []
        gt_info = {}
        for file in gt_files:
            data_info = file_utils.read_data(file, split=" ")
            name = [info[0] for info in data_info]
            boxes = [info[1:] for info in data_info]
            class_name += name
            self.build_data_info(gt_info, os.path.basename(file), boxes, name)
        class_name = list(set(class_name))
        class_name = sorted(class_name)
        print(class_name)
        self.evaluate(gt_info, dt_info, class_name, plot=True)


if __name__ == "__main__":
    val = MapMetric()
    val.evaluate_for_json()
    # val.evaluate_for_txt()



# -*- coding: utf-8 -*-
"""
# --------------------------------------------------------
# @Project: FaceRecognitionTest
# @Author : panjq
# @E-mail : pan_jinquan@163.com
# @Date   : 2020-03-19 09:19:18
# --------------------------------------------------------
"""

import numpy as np
from . import iou


class MatchingBBoxes(object):
    """bboxes matching"""

    @staticmethod
    def get_matching_map(pre_bboxes, true_bboxes, iou_threshold=0., axis=1):
        """
        输入IOU矩阵,根据IOU进行位置匹配,返回满足阈值的匹配的映射关系表
        :param pre_bboxes,
        :param true_bboxes
        :param iou_threshold: 0:匹配所有true_bboxes,
                           1e-5:匹配iou>= 1e-5所有true_bboxes
        :param axis:
        :return: max_iou  : 匹配分数,<class 'tuple'>: ((len(true_data),)
                 pt_map: 匹配的映射关系表,<class 'tuple'>: (len(true_data), 2)
                            index_max[:, 0]: pred_index
                            index_max[:, 1]: true_index = [0,1,2,...,len(true_data)]是下标从0开始的递增序列
        """
        # 获得IOU矩阵iou_mat:其中纵轴表示true_data,横轴表示pre_data
        iou_mat = iou.get_iou_mat(true_bboxes, pre_bboxes)  # <class 'tuple'>: (len(true_bboxes), len(pre_bboxes))
        pred_index = np.argmax(iou_mat, axis=axis)  # max_index,<class 'tuple'>: (len(true_bboxes),)
        true_index = np.arange(0, len(pred_index))
        pt_map = np.asarray((pred_index, true_index)).T
        max_iou = np.max(iou_mat, axis=axis)
        index = max_iou >= iou_threshold
        max_iou = max_iou[index]
        pt_map = pt_map[index]
        return max_iou, pt_map

    @staticmethod
    def get_matching_index(pred_bboxes, true_bboxes, iou_threshold=0):
        """
        BUG:当iou_threshold为0时,pred_bboxes, true_bboxes维度不变,
            但当iou_threshold>0,pred_bboxes, true_bboxes输出维度可能变小,
        :param pred_bboxes:
        :param true_bboxes:
        :param iou_threshold:
        :return: 返回匹配关系:
                    pred_index: 匹配后对应pred_bboxes的index
                    true_index: 匹配后对应true_bboxes的index,[0,1,2,...,len(true_data)]是下标从0开始的递增序列        
        """
        if not isinstance(pred_bboxes, np.ndarray):
            pred_bboxes = np.asarray(pred_bboxes)
        if not isinstance(true_bboxes, np.ndarray):
            true_bboxes = np.asarray(true_bboxes)
        max_iou, pt_map = MatchingBBoxes.get_matching_map(pred_bboxes, true_bboxes, iou_threshold)
        pred_index = pt_map[:, 0]
        true_index = pt_map[:, 1]
        return pred_index, true_index

    @staticmethod
    def bboxes_matching(pred_bboxes, true_bboxes, iou_threshold=0):
        """
        BUG:当iou_threshold为0时,pred_bboxes, true_bboxes维度不变,
            但当iou_threshold>0,pred_bboxes, true_bboxes输出维度可能变小,
        :param pred_bboxes:
        :param true_bboxes:
        :param iou_threshold:
        :return:
        """
        if not isinstance(pred_bboxes, np.ndarray):
            pred_bboxes = np.asarray(pred_bboxes)
        if not isinstance(true_bboxes, np.ndarray):
            true_bboxes = np.asarray(true_bboxes)
        max_iou, pt_map = MatchingBBoxes.get_matching_map(pred_bboxes, true_bboxes, iou_threshold)
        pred_index = pt_map[:, 0]
        true_index = pt_map[:, 1]
        pred_bboxes = pred_bboxes[pred_index]
        true_bboxes = true_bboxes[true_index]
        return pred_bboxes, true_bboxes

    @staticmethod
    def bboxes_id_matching(pred_bboxes, true_bboxes, pred_id, iou_threshold=0.):
        """
        BUG:当iou_threshold为0时,face_bboxes, body_bboxes维度不变,
            但当iou_threshold>0,face_bboxes, body_bboxes输出维度可能变小,
        :param pred_bboxes:
        :param pred_id
        :param true_bboxes:
        :param iou_threshold:
        :return:
        """
        if not isinstance(pred_bboxes, np.ndarray):
            pred_bboxes = np.asarray(pred_bboxes)
        if not isinstance(true_bboxes, np.ndarray):
            true_bboxes = np.asarray(true_bboxes)
        if not isinstance(pred_id, np.ndarray):
            pred_id = np.asarray(pred_id)
        max_iou, pt_map = MatchingBBoxes.get_matching_map(pred_bboxes, true_bboxes, iou_threshold)
        pred_index = pt_map[:, 0]
        true_index = pt_map[:, 1]
        pred_bboxes = pred_bboxes[pred_index]
        true_bboxes = true_bboxes[true_index]
        pred_id = pred_id[pred_index]
        return pred_bboxes, true_bboxes, pred_id

    @staticmethod
    def face_body_matching(face_bboxes, body_bboxes, body_keys, face_keys, iou_threshold=1e-5, **kwargs):
        """
        实现人脸/人体bboes匹配算法(PS:将人体框body_bboxes匹配到人脸框face_bboxes,因此不会修改face_bboxes顺序)
        :param face_bboxes:<np.ndarray, list>人脸框(被匹配的bboxes)
        :param body_bboxes:<np.ndarray, list>人体框(待匹配的bboxes)
        :param body_keys:<set>与body_bboxes绑定的信息的key
        :param face_keys:<set>与face_bboxes绑定的信息的key
        :param body_keys:<np.ndarray, list>人体框(待匹配的bboxes)
        :param iou_threshold:<float>,人脸/人体框匹配阈值,即IOU阈值最低分数
        :param kwargs:<dict, optional>跟人脸框对应的相关数据,长度必须与face_bboxes一样,且跟face_bboxes是一一对应的关系
                      比如一个人脸框,对应一个人脸识别ID,或者对应一个表情识别标签.格式如
                      kwargs= {"emotion": {"labels": [], "scores": []},
                               "others":  {"labels": []},
                                ...
                                ...
                                ...
                            }
        :return:返回匹配后的数据
        """
        if not isinstance(face_bboxes, np.ndarray):
            face_bboxes = np.asarray(face_bboxes)
        if not isinstance(body_bboxes, np.ndarray):
            body_bboxes = np.asarray(body_bboxes)
        # 匹配人体框到人脸框,获得index对应关系
        body_index, face_index = MatchingBBoxes.get_matching_index(body_bboxes,
                                                                   face_bboxes,
                                                                   iou_threshold)
        face_bboxes = face_bboxes[face_index]
        body_bboxes = body_bboxes[body_index]

        # 脸框依赖的数据,需要根据face_index进行调整,避免丢失对应关系
        # 人体框依赖的数据,需要根据body_index进行调整,避免丢失对应关系
        out_kwargs = {}
        for name, dict_val in kwargs.items():
            out_kwargs[name] = {}
            for k, v in dict_val.items():
                if not isinstance(v, np.ndarray):
                    v = np.asarray(v)
                # 按照数据依赖不同（人体框或人脸框）进行顺序调整
                if name in body_keys:
                    out_kwargs[name][k] = v[body_index]
                elif name in face_keys:
                    out_kwargs[name][k] = v[face_index]

        return face_bboxes, body_bboxes, out_kwargs



# -*- coding: utf-8 -*-
"""
    @Author : Pan
    @E-mail : 390737991@qq.com
    @Date   : 2022-12-30 10:40:48
    @Brief  :
"""
import os, sys

sys.path.append(os.getcwd())
import onnx
import onnxruntime
import numpy as np


class ONNXEngine():
    def __init__(self, model_file, use_gpu=False):
        """
        :param model_file:
        :param use_gpu:
        pip install onnxruntime-gpu -i https://pypi.tuna.tsinghua.edu.cn/simple
        """
        available_providers = onnxruntime.get_available_providers()
        sess_options = onnxruntime.SessionOptions()
        if use_gpu:
            self.providers = ['CUDAExecutionProvider']
        else:
            self.providers = ['CPUExecutionProvider']
        self.onnx_session = onnxruntime.InferenceSession(model_file, providers=self.providers,
                                                         sess_options=sess_options)
        self.device = onnxruntime.get_device()
        self.inp_names = self.get_inp_names(self.onnx_session)
        self.out_names = self.get_out_names(self.onnx_session)
        print("available_providers:{},use device:{}".format(available_providers, self.device))
        print("inp_names          :{}".format(self.inp_names))
        print("out_names          :{}".format(self.out_names))
        print("model_file          :{}".format(model_file))
        print('-----------' * 5, flush=True)

    def get_out_names(self, onnx_session):
        """
        output_name = onnx_session.get_outputs()[0].name
        :param onnx_session:
        :return:
        """
        output_name = []
        for node in onnx_session.get_outputs():
            output_name.append(node.name)
        return output_name

    def get_inp_names(self, onnx_session):
        """
        input_name = onnx_session.get_inputs()[0].name
        :param onnx_session:
        :return:
        """
        input_name = []
        for node in onnx_session.get_inputs():
            input_name.append(node.name)
        return input_name

    def get_inp_feed(self, input_name, image_tensor):
        """
        input_feed={self.input_name: image_tensor}
        :param input_name:
        :param image_tensor:
        :return:
        """
        input_feed = {}
        for name in input_name:
            input_feed[name] = image_tensor
        return input_feed

    def __call__(self, image_tensor):
        """
        num_bboxes=1500
        num_class=1
        :param image_tensor: shape
        :return: outputs[0]
        """
        outputs = self.forward(image_tensor)
        return outputs

    def forward(self, inp_tensor):
        """
        image_tensor = image.transpose(2, 0, 1)
        image_tensor = image_tensor[np.newaxis, :]
        onnx_session.run([output_name], {input_name: x})
        :param inp_tensor:
        :return:
        """
        # 输入数据的类型必须与模型一致,以下三种写法都是可以的
        # scores, boxes = self.onnx_session.run(None, {self.input_name: image_tensor})
        # scores, boxes = self.onnx_session.run(self.output_name, input_feed={self.input_name: image_tensor})
        inp_feed = self.get_inp_feed(self.inp_names, inp_tensor)
        out_tensor = self.onnx_session.run(self.out_names, input_feed=inp_feed)
        return out_tensor

    def performance(self, inputs, iterate=10):
        from pybaseutils import time_utils
        outputs = self.forward(inputs)
        for i in range(iterate):
            with time_utils.Performance() as p:
                outputs = self.forward(inputs)
        return outputs


if __name__ == "__main__":
    model_file = "libs/best.onnx"
    batch_size = 1
    num_classes = 4
    input_size = [320, 320]
    inputs = np.random.random(size=(batch_size, 3, input_size[1], input_size[0]))
    inputs = np.asarray(inputs, dtype=np.float32)
    model = ONNXEngine(model_file, use_gpu=False)
    model.performance(inputs)
    print("----")



# -*-coding: utf-8 -*-
"""
    @Author : PKing
    @E-mail : 390737991@qq.com
    @Date   : 2022-12-26 15:37:56
    @Brief  : https://github.com/Tencent/ncnn/wiki/use-ncnn-with-pytorch-or-onnx
              https://github.com/pnnx/pnnx
"""

import os
import torch.onnx
import pnnx


def convert2ncnn(model, input_shape, input_names=['input'], output_names=['output'],
                 out_file="", fp16=True, use_prune=False, sparsity=0.2, device="cuda:0"):
    """
    https://github.com/Tencent/ncnn/wiki/use-ncnn-with-pytorch-or-onnx
    :param model: Pytorch 模型
    :param input_shape: 输入维度(B, C, H, W)
    :param input_names: 输入节点名称
    :param output_names: 输出节点名称
    :param out_file: 输出模型文件     out_file = model_file.replace(".pth", ".jit").replace(".pt", ".jit")
    :param use_prune: 是否对模型进行剪枝
    :param sparsity: 对模型进行剪枝的稀疏度
    :param device: 运行设备
    :return:
    """
    if out_file:
        output = os.path.dirname(out_file)
    else:
        output = "./output"
        out_file = os.path.join(output, "model.jit")
    if not os.path.exists(output): os.makedirs(output)
    B, C, H, W = input_shape
    if use_prune:
        from basetrainer.pruning import nni_pruning
        model = nni_pruning.model_pruning(model,
                                          input_size=[8, 3, H, W],
                                          sparsity=sparsity,
                                          reuse=True,
                                          output_prune=os.path.join(output, "prune"),
                                          device=device)
    model = model.to(device)
    model.eval()
    inputs = torch.randn(B, C, H, W).to(device)
    model_ = pnnx.export(model, out_file, inputs=inputs, fp16=fp16)
    print("save ncnn modle file:{}".format(os.path.dirname(out_file)))
    return output


if __name__ == "__main__":
    pass
